SAL_SyncComm = {};
SAL_SyncComm.Version = 1;

SAL_SyncComm.STATUS_SENDING = "SENDING";
SAL_SyncComm.STATUS_RECEIVING = "RECEIVING";
SAL_SyncComm.STATUS_SENTREQUEST = "SENTREQUEST";
SAL_SyncComm.STATUS_GOTREQUEST = "GOTREQUEST";

--------- Node encoding ----------------

local function decodeNum(str) -- LSB first
  --print("decoding", str);
  local values = {strbyte(str,1,-1)};
  local base = 1;
  local value = 0;
  for i,v in ipairs(values) do
    value = value + (v - 1)*base;
    base = base * 255;
  end
  return value;
end

local function encodeNum(num, size) --LSB first
  local filler = strrep(strchar(1), size);
  local l = {};
  while (num > 0) do
    local val = mod(num, 255);
    l[#l + 1] = val + 1;
    num = floor(num / 255);
  end
  local str = strsub(filler .. strchar(unpack(l)), -size);
  return str;
end

local function encodeNode(x, y, count)
  local xStr = encodeNum(floor(x * 254 * 254), 2);
  local yStr = encodeNum(floor(y * 254 * 254), 2);
  local cStr = encodeNum(min(count, 250), 1);
  local str = xStr .. yStr .. cStr;
  --print("Encoding Node", count, x, y, strlen(str), str);
  return str;
end

local function decodeNode(str)
  local x = decodeNum(strsub(str, 1, 2)) /254/254; -- read first 2 chars
  local y = decodeNum(strsub(str, 3, 4)) /254/254; -- read next 2 chars
  local count = decodeNum(strsub(str, 5, 5)); -- read 1 for count

  --print("Decoding node", count, x, y, strlen(str), str);
  return x, y, count;
end

----------------- Message sending -----------------------

local function sendMsg(target, message)
	--print("Sending message to", target, message)
	SendAddonMessage("SAL", message, "WHISPER", target);
end

local function sendCmd(target, cmd, ...)
	--print("Sending command to", target, cmd, ...)
	local args = {...};
	local msg = cmd;
	for i,v in ipairs(args) do
		if (type(v) == "string") then
			msg = msg .. " '" .. v .. "'";
		else
			msg = msg .. " " .. v;
		end
	end
	sendMsg(target, msg);
end

------------------- Send commands -----------------------------

function SAL_SyncComm:SyncRequest(to, mapCount, nodeCount, source)
	if (SAL_SyncComm.Current == nil) then
		SAL_SyncComm.Current = {otherSide=strlower(to), status=SAL_SyncComm.STATUS_SENTREQUEST, time=GetTime(), mapCount=mapCount, nodeCount=nodeCount};
		--print("Current set");
		sendCmd(to, "SyncRequest", SAL_SyncComm.Version, mapCount, nodeCount, source);		
	end
end

function SAL_SyncComm:SyncAbort(to, message)
	sendCmd(to, "SyncAbort", message);
end

function SAL_SyncComm:ReplySync(to, message)
	if (SAL_SyncComm.Current ~= nil) then
		SAL_SyncComm.Current.time = GetTime();
		SAL_SyncComm.Current.status = SAL_SyncComm.STATUS_RECEIVING;
		sendCmd(to, "ReplySync", "OK");
	end
end

function SAL_SyncComm:Data(to, mapName, nodes)
	local msg = "Data '" .. mapName .. "'";
	for i,v in ipairs(nodes) do
		if (strlen(msg) > 245) then
			sendMsg(to, msg);
			msg = "Data '" .. mapName .. "'";
		end
		
		msg = msg .. encodeNode(v[1], v[2], v.n);
	end
	sendMsg(to, msg);
	SAL_SyncComm.Current.time = GetTime();
end

function SAL_SyncComm:DataEnd(to)
	sendCmd(to, "DataEnd");
	SAL_SyncComm.Current = nil;
end

----------------- Handlers -----------------------------

local function SyncRequestHandler(sender, version, mapCount, nodeCount, source)
	--print("got sync req");
	if (SAL_SyncComm.Current ~= nil) then
		SAL_SyncComm:SyncAbort(sender, "Receiver is busy with another transaction");
	elseif (SAL_SyncComm.Version ~= version) then
		SAL_SyncComm:SyncAbort(sender, string.format("Version mismatch, sender %i, receiver %i", version, SAL_SyncComm.Version));
		SAL_SyncComm.Current = nil;
	elseif (UnitIsDND("player")) then
		SAL_SyncComm:SyncAbort(sender, "Target flagged as DND");
	else
		SAL_SyncComm.Current = {otherSide=strlower(sender), status=SAL_SyncComm.STATUS_GOTREQUEST, time=GetTime(), mapCount=mapCount, nodeCount=nodeCount};
		SAL_SyncComm.RequestHandler(sender, mapCount, nodeCount, source);
	end
end

local function SyncAbortHandler(sender, message)
	--print("got abort", SAL_SyncComm.Current, SAL_SyncComm.Current.otherSide, sender);
	
	if (SAL_SyncComm.Current ~= nil and SAL_SyncComm.Current.otherSide == sender) then
		--print("Aborting");
		SAL_SyncComm.Current = nil;
		SAL_SyncComm.AbortHandler(message);
	end
end

local function ReplySyncHandler(sender, message)
	--print("got reply");
	if (SAL_SyncComm.Current ~= nil 
		and SAL_SyncComm.Current.otherSide == sender
		and SAL_SyncComm.Current.status == SAL_SyncComm.STATUS_SENTREQUEST) then
		if (message == "OK") then
			SAL_SyncComm.Current.status = SAL_SyncComm.STATUS_SENDING;
			SAL_SyncComm.Current.time = GetTime();
			SAL_SyncComm.ReplyHandler(true);
		else
			SAL_SyncComm.Current = nil;
			SAL_SyncComm.ReplyHandler(false, message);
		end
	end
end

local function DataHandler(sender, map, coords)
	--print("got data");
	if (SAL_SyncComm.Current ~= nil
		and SAL_SyncComm.Current.otherSide == sender
		and SAL_SyncComm.Current.status == SAL_SyncComm.STATUS_RECEIVING) then
		SAL_SyncComm.Current.time = GetTime();
		SAL_SyncComm.DataHandler(map, coords);
	end
end

local function DataEndHandler(sender)
	if (SAL_SyncComm.Current ~= nil
		and SAL_SyncComm.Current.otherSide == sender
		and SAL_SyncComm.Current.status == SAL_SyncComm.STATUS_RECEIVING) then
		SAL_SyncComm.Current = nil;
		SAL_SyncComm.DataEndHandler();
	end
end

---------------- Message receiving --------------------------

local function incommingCmd(sender, cmd, args)
	--print("Got command from", sender, cmd, unpack(args));
	local f = SAL_SyncComm.messageRouter[cmd];
	if (f ~= nil) then
		f(sender, unpack(args));
	end
end

local function ReadToNext(str, char)
	local len = strlen(str);
	
	for i = 1,len do
		local c = strsub(str, i, i);
		if (c == char) then
			return strsub(str, 1, i), strsub(str, i+1);
		end
	end
	return str, "";
end

local function ParamSplit(str)
	local l = {};

	while(str ~= nil and strlen(str) > 0) do
		local x, y;
		if (strsub(str, 1, 1) == "'") then
			x, y = ReadToNext(strsub(str, 2), "'");
			x = strtrim(x, "'");
			y = strsub(y, 2);
		else
			x, y = ReadToNext(str, " ");
			x = tonumber(x);
		end
		l[#l+1] = x;
		str = y;
	end
	
	return l;
end

local function incommingMsg(sender, cmd, msg)
	--print("got message", cmd, msg);
	local args;
	if (cmd == "Data") then
		local _, map, coordsStr = strsplit("'", msg, 3);
		local coords = {};
		for i = 1, strlen(coordsStr), 5 do
			local coordStr = strsub(coordsStr, i, i + 4);
			local x, y, n = decodeNode(coordStr);
			local coord = {x,y};
			coord.n = n;
			coords[#coords + 1] = coord;
		end
		args = {map, coords};
	else
		args = ParamSplit(msg);
	end
	incommingCmd(sender, cmd, args);
end

function SAL_incommingChat(prefix, message, channel, sender)
	if (prefix == "SAL") then
		--print("Incomming chat event", prefix, message, channel, sender);
		local cmd, msg = strsplit(" ", message, 2);
		incommingMsg(strlower(sender), cmd, msg);
	end
end

------------------------------------------------

function SAL_SyncComm:Cleanup()
	if (SAL_SyncComm.Current ~= nil) then
		if (SAL_SyncComm.Current.Status == SAL_SyncComm.STATUS_RECEIVING
			and GetTime() > SAL_SyncComm.Current.time + 5) then
			-- Timeout while sending
			SAL_SyncComm:SyncAbort(SAL_SyncComm.Current.otherSide, "Timeout");
			SAL_SyncComm.Current = nil;
			SAL_SyncComm.ErrorHandler("Timeout");
		elseif(SAL_SyncComm.Current.Status == SAL_SyncComm.STATUS_SENTREQUEST
			and GetTime() > SAL_SyncComm.Current.time + 60) then
			-- Timeout while waiting for a request
			SAL_SyncComm:SyncAbort(SAL_SyncComm.Current.otherSide, "Timeout");
			SAL_SyncComm.Current = nil;
			SAL_SyncComm.ErrorHandler("Timeout");
		end
	end
end

function SAL_SyncComm:Init(Request, Abort, Reply, Data, DataEnd, Error)
	SAL_SyncComm.messageRouter = {};
	SAL_SyncComm.messageRouter["SyncRequest"] = SyncRequestHandler;
	SAL_SyncComm.messageRouter["SyncAbort"] = SyncAbortHandler;
	SAL_SyncComm.messageRouter["ReplySync"] = ReplySyncHandler;
	SAL_SyncComm.messageRouter["Data"] = DataHandler;
	SAL_SyncComm.messageRouter["DataEnd"] = DataEndHandler;
	
	SAL_SyncComm.RequestHandler = Request;
	SAL_SyncComm.AbortHandler = Abort;
	SAL_SyncComm.ReplyHandler = Reply;
	SAL_SyncComm.DataHandler = Data;
	SAL_SyncComm.DataEndHandler = DataEnd;
	SAL_SyncComm.ErrorHandler = Error;
	
	-- Call cleanup from frame update
end
